import os
import sys
import json
import logging
import threading
from copy import copy
from datetime import date, timedelta

import boto3
import requests
from six.moves.queue import Queue
from six import iteritems, iterkeys
from sentinel_s3.crawler import get_products_metadata_path, get_product_metadata_path
from sentinel_s3.converter import metadata_to_dict, tile_metadata

# Python 2 comptability
try:
    JSONDecodeError = json.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError


logger = logging.getLogger('sentinel.meta.s3')

bucket_name = os.getenv('BUCKETNAME', 'sentinel-metadata')
s3 = boto3.resource('s3')
bucket = s3.Bucket(bucket_name)


def mkdirp(path):

    if not os.path.exists(path):
        os.makedirs(path)


def file_writer(product_dir, metadata):
    mkdirp(product_dir)

    f = open(os.path.join(product_dir, metadata['tile_name'] + '.json'), 'w')
    f.write(json.dumps(metadata))
    f.close()


def s3_writer(product_dir, metadata):
    # make sure product_dir doesn't start with slash (/) or dot (.)
    if product_dir.startswith('.'):
        product_dir = product_dir[1:]

    if product_dir.startswith('/'):
        product_dir = product_dir[1:]

    key = os.path.join(product_dir, metadata['tile_name'] + '.json')
    s3.Object(bucket_name, key).put(json.dumps(metadata))
    object_acl = s3.ObjectAcl(bucket_name, key)
    object_acl.put(ACL='public-read')


def product_metadata(product, dst_folder, counter=None, writers=[file_writer], geometry_check=None):
    """ Extract metadata for a specific product """

    if not counter:
        counter = {
            'products': 0,
            'saved_tiles': 0,
            'skipped_tiles': 0,
            'skipped_tiles_paths': []
        }

    s3_url = 'http://sentinel-s2-l1c.s3.amazonaws.com'

    product_meta_link = '{0}/{1}'.format(s3_url, product['metadata'])
    product_info = requests.get(product_meta_link, stream=True)
    product_metadata = metadata_to_dict(product_info.raw)
    product_metadata['product_meta_link'] = product_meta_link

    counter['products'] += 1

    for tile in product['tiles']:
        tile_info = requests.get('{0}/{1}'.format(s3_url, tile))
        try:
            metadata = tile_metadata(tile_info.json(), copy(product_metadata), geometry_check)

            for w in writers:
                w(dst_folder, metadata)

            logger.info('Saving to disk: %s' % metadata['tile_name'])
            counter['saved_tiles'] += 1
        except JSONDecodeError:
            logger.warning('Tile: %s was not found and skipped' % tile)
            counter['skipped_tiles'] += 1
            counter['skipped_tiles_paths'].append(tile)

    return counter


def single_metadata(product_name, dst_folder, writers=[file_writer], geometry_check=None):

    product_list = get_product_metadata_path(product_name)
    return product_metadata(product_list[product_name], dst_folder, writers=writers, geometry_check=geometry_check)


def daily_metadata(year, month, day, dst_folder, writers=[file_writer], geometry_check=None,
                   num_worker_threads=1):
    """ Extra metadata for all products in a specific date """

    threaded = False

    counter = {
        'products': 0,
        'saved_tiles': 0,
        'skipped_tiles': 0,
        'skipped_tiles_paths': []
    }

    if num_worker_threads > 1:
        threaded = True
        queue = Queue()

    # create folders
    year_dir = os.path.join(dst_folder, str(year))
    month_dir = os.path.join(year_dir, str(month))
    day_dir = os.path.join(month_dir, str(day))

    product_list = get_products_metadata_path(year, month, day)

    logger.info('There are %s products in %s-%s-%s' % (len(list(iterkeys(product_list))),
                                                       year, month, day))

    for name, product in iteritems(product_list):
        product_dir = os.path.join(day_dir, name)

        if threaded:
            queue.put([product, product_dir, counter, writers, geometry_check])
        else:
            counter = product_metadata(product, product_dir, counter, writers, geometry_check)

    if threaded:
        def worker():
            while not queue.empty():
                args = queue.get()
                try:
                    product_metadata(*args)
                except Exception:
                    exc = sys.exc_info()
                    logger.error('%s tile skipped due to error: %s' % (threading.current_thread().name,
                                                                       exc[1].__str__()))
                    args[2]['skipped_tiles'] += 1
                queue.task_done()

        threads = []
        for i in range(num_worker_threads):
            t = threading.Thread(target=worker)
            t.start()
            threads.append(t)

        queue.join()

    return counter


def range_metadata(start, end, dst_folder, num_worker_threads=0, writers=[file_writer], geometry_check=None):
    """ Extra metadata for all products in a date range """

    assert isinstance(start, date)
    assert isinstance(end, date)

    delta = end - start

    dates = []

    for i in range(delta.days + 1):
        dates.append(start + timedelta(days=i))

    days = len(dates)

    total_counter = {
        'days': days,
        'products': 0,
        'saved_tiles': 0,
        'skipped_tiles': 0,
        'skipped_tiles_paths': []
    }

    def update_counter(counter):
        for key in iterkeys(total_counter):
            if key in counter:
                total_counter[key] += counter[key]

    for d in dates:
        logger.info('Getting metadata of {0}-{1}-{2}'.format(d.year, d.month, d.day))
        update_counter(daily_metadata(d.year, d.month, d.day, dst_folder, writers, geometry_check,
                                      num_worker_threads))

    return total_counter



import json
import datetime

import boto3

bucket_name = 'sentinel-s2-l1c'
s3 = boto3.resource('s3', region_name='eu-central-1')
bucket = s3.Bucket(bucket_name)


def get_tile_metadata_path(path):

    meta_obj = s3.Object(bucket_name, path)

    meta = json.loads(meta_obj.get()['Body'].read().decode())

    paths = []
    for tile in meta['tiles']:
        paths.append(tile['path'] + '/tileInfo.json')

    return paths


def get_product_metadata_path(product_name):
    """ gets a single products metadata """

    string_date = product_name.split('_')[-1]
    date = datetime.datetime.strptime(string_date, '%Y%m%dT%H%M%S')
    path = 'products/{0}/{1}/{2}/{3}'.format(date.year, date.month, date.day, product_name)

    return {
        product_name: {
            'metadata': '{0}/{1}'.format(path, 'metadata.xml'),
            'tiles': get_tile_metadata_path('{0}/{1}'.format(path, 'productInfo.json'))
        }
    }


def get_products_metadata_path(year, month, day):
    """ Get paths to multiple products metadata """

    products = {}
    path = 'products/{0}/{1}/{2}/'.format(year, month, day)
    for key in bucket.objects.filter(Prefix=path):
        product_path = key.key.replace(path, '').split('/')
        name = product_path[0]
        if name not in products:
            products[name] = {}

        if product_path[1] == 'metadata.xml':
            products[name]['metadata'] = key.key

        if product_path[1] == 'productInfo.json':
            products[name]['tiles'] = get_tile_metadata_path(key.key)

    return products



import re
import os
import errno
import shutil
import logging
import threading
from tempfile import mkdtemp
from collections import OrderedDict
import xml.etree.cElementTree as etree

import rasterio
import requests
from wordpad import pad
from six import iteritems
from pyproj import Proj, transform
from rasterio.features import shapes
from shapely.ops import cascaded_union
from shapely.geometry import mapping, Polygon

logger = logging.getLogger('sentinel.meta.s3')
s3_url = 'http://sentinel-s2-l1c.s3.amazonaws.com'


def epsg_code(geojson):
    """ get the espg code from the crs system """

    if isinstance(geojson, dict):
        if 'crs' in geojson:
            urn = geojson['crs']['properties']['name'].split(':')
            if 'EPSG' in urn:
                try:
                    return int(urn[-1])
                except (TypeError, ValueError):
                    return None

    return None


def convert_coordinates(coords, origin, wgs84):
    """ Convert coordinates from one crs to another """
    if isinstance(coords, list) or isinstance(coords, tuple):
        try:
            if isinstance(coords[0], list) or isinstance(coords[0], tuple):
                return [convert_coordinates(list(c), origin, wgs84) for c in coords]
            elif isinstance(coords[0], float):
                return list(transform(origin, wgs84, *coords))
        except IndexError:
            pass

    return None


def to_latlon(geojson, origin_espg=None):
    """
    Convert a given geojson to wgs84. The original epsg must be included insde the crs
    tag of geojson
    """

    if isinstance(geojson, dict):

        # get epsg code:
        if origin_espg:
            code = origin_espg
        else:
            code = epsg_code(geojson)
        if code:
            origin = Proj(init='epsg:%s' % code)
            wgs84 = Proj(init='epsg:4326')

            new_coords = convert_coordinates(geojson['coordinates'], origin, wgs84)
            if new_coords:
                geojson['coordinates'] = new_coords
                if 'crs' not in geojson:
                    geojson['crs'] = {
                        'type': 'name',
                        'properties': {
                            'name': 'urn:ogc:def:crs:EPSG:8.9:4326'
                        }
                    }
                else:
                    geojson['crs']['properties']['name'] = 'urn:ogc:def:crs:EPSG:8.9:4326'

    return geojson


def camelcase_underscore(name):
    """ Convert camelcase names to underscore """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def get_tiles_list(element):
    """
    Returns the list of all tile names from Product_Organisation element
    in metadata.xml
    """

    tiles = {}

    for el in element:
        g = el.findall('.//Granules')[0]
        name = g.attrib['granuleIdentifier']

        name_parts = name.split('_')
        mgs = name_parts[-2]
        tiles[mgs] = name

    return tiles


def metadata_to_dict(metadata):
    """ Looks at metadata.xml file of sentinel product and extract useful keys
    Returns a python dict """

    tree = etree.parse(metadata)
    root = tree.getroot()

    meta = OrderedDict()

    keys = [
        'SPACECRAFT_NAME',
        'PRODUCT_STOP_TIME',
        'Cloud_Coverage_Assessment',
        'PROCESSING_LEVEL',
        'PRODUCT_TYPE',
        'PROCESSING_BASELINE',
        'SENSING_ORBIT_NUMBER',
        'SENSING_ORBIT_DIRECTION',
        'PRODUCT_FORMAT',
    ]

    # grab important keys from the file
    for key in keys:
        try:
            meta[key.lower()] = root.findall('.//' + key)[0].text
        except IndexError:
            meta[key.lower()] = None

    meta['product_cloud_coverage_assessment'] = float(meta.pop('cloud_coverage_assessment'))

    meta['sensing_orbit_number'] = int(meta['sensing_orbit_number'])

    # get tile list
    meta['tiles'] = get_tiles_list(root.findall('.//Product_Organisation')[0])

    # get available bands
    bands = root.findall('.//Band_List')[0]
    meta['band_list'] = []
    for b in bands:
        band = b.text.replace('B', '')
        if len(band) == 1:
            band = 'B' + pad(band, 2)
        else:
            band = b.text
        meta['band_list'].append(band)

    return meta


def get_tile_geometry(path, origin_espg, tolerance=500):
    """ Calculate the data and tile geometry for sentinel-2 tiles """

    with rasterio.open(path) as src:

        # Get tile geometry
        b = src.bounds
        tile_shape = Polygon([(b[0], b[1]), (b[2], b[1]), (b[2], b[3]), (b[0], b[3]), (b[0], b[1])])
        tile_geojson = mapping(tile_shape)

        # read first band of the image
        image = src.read(1)

        # create a mask of zero values
        mask = image == 0.

        # generate shapes of the mask
        novalue_shape = shapes(image, mask=mask, transform=src.affine)

        # generate polygons using shapely
        novalue_shape = [Polygon(s['coordinates'][0]) for (s, v) in novalue_shape]

        if novalue_shape:

            # Make sure polygons are united
            # also simplify the resulting polygon
            union = cascaded_union(novalue_shape)

            # generates a geojson
            data_shape = tile_shape.difference(union)

            # If there are multipolygons, select the largest one
            if data_shape.geom_type == 'MultiPolygon':
                areas = {p.area: i for i, p in enumerate(data_shape)}
                largest = max(areas.keys())
                data_shape = data_shape[areas[largest]]

            # if the polygon has interior rings, remove them
            if list(data_shape.interiors):
                data_shape = Polygon(data_shape.exterior.coords)

            data_shape = data_shape.simplify(tolerance, preserve_topology=False)
            data_geojson = mapping(data_shape)

        else:
            data_geojson = tile_geojson

        # convert cooridnates to degrees
        return (to_latlon(tile_geojson, origin_espg), to_latlon(data_geojson, origin_espg))


def get_tile_geometry_from_s3(meta):

    # create a temp folder
    tmp_folder = mkdtemp()
    f = os.path.join(tmp_folder, 'B01.jp2')

    # donwload B01
    r = requests.get('{0}/{1}/B01.jp2'.format(s3_url, meta['path']), stream=True)
    chunk_size = 1024

    with open(f, 'wb') as fd:
        for chunk in r.iter_content(chunk_size):
            fd.write(chunk)

    (meta['tile_geometry'],
     meta['tile_data_geometry']) = get_tile_geometry(f, epsg_code(meta['tile_geometry']))
    meta['tile_origin'] = to_latlon(meta['tile_origin'])

    # remove temp folder
    try:
        shutil.rmtree(tmp_folder)
    except OSError as exc:
        if exc.errno != errno.ENOENT:
            raise

    return meta


def tile_metadata(tile, product, geometry_check=None):
    """ Generate metadata for a given tile

    - geometry_check is a function the determines whether to calculate the geometry by downloading
    B01 and override provided geometry in tilejson. The meta object is passed to this function.
    The function return a True or False response.
    """

    grid = 'T{0}{1}{2}'.format(pad(tile['utmZone'], 2), tile['latitudeBand'], tile['gridSquare'])

    meta = OrderedDict({
        'tile_name': product['tiles'][grid]
    })

    logger.info('%s Processing tile %s' % (threading.current_thread().name, tile['path']))

    meta['date'] = tile['timestamp'].split('T')[0]

    meta['thumbnail'] = '{1}/{0}/preview.jp2'.format(tile['path'], s3_url)

    # remove unnecessary keys
    product.pop('tiles')
    tile.pop('datastrip')
    bands = product.pop('band_list')

    for k, v in iteritems(tile):
        meta[camelcase_underscore(k)] = v

    meta.update(product)

    # construct download links
    links = ['{2}/{0}/{1}.jp2'.format(meta['path'], b, s3_url) for b in bands]

    meta['download_links'] = {
        'aws_s3': links
    }

    meta['original_tile_meta'] = '{0}/{1}/tileInfo.json'.format(s3_url, meta['path'])

    def internal_latlon(meta):
        keys = ['tile_origin', 'tile_geometry', 'tile_data_geometry']
        for key in keys:
            if key in meta:
                meta[key] = to_latlon(meta[key])
        return meta

    # change coordinates to wsg4 degrees
    if geometry_check:
        if geometry_check(meta):
            meta = get_tile_geometry_from_s3(meta)
        else:
            meta = internal_latlon(meta)
    else:
        meta = internal_latlon(meta)

    # rename path key to aws_path
    meta['aws_path'] = meta.pop('path')

    return meta



from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = '0.6.1'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if 'git+' not in x]

setup(
    name='sentinel-s3',
    version=__version__,
    description='Python libraries for extracting Sentinel-2\'s metadata from Amazon S3',
    long_description=long_description,
    url='https://github.com/developmentseed/sentinel-s3',
    download_url='https://github.com/developmentseed/sentinel-s3/tarball/' + __version__,
    license='CC0',
    classifiers=[
        'Framework :: Pytest',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: Freeware',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author='Alireza J (scisco)',
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='alireza@developmentseed.org',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)
